from .base import RecurringDateTime
from .base import next_date_of_same_weekday, _get_isoweekday_difference
from datetime import datetime
from icemac.recurrence.i18n import _
import gocept.month
import grokcore.component as grok
import math
import zope.cachedescriptors.property


def recurrences_of_weekday_in_month(date, month):
    """Return number of recurrences of weekday of `date` in `month`."""
    minus_days = _get_isoweekday_difference(month.firstOfMonth(), date)
    return int(math.ceil((month.lastOfMonth().day - minus_days) / 7.0))


class SameNthWeekdayInMonthBase(RecurringDateTime):
    """Base class for recurrings on the same day nth weekday in month."""

    grok.baseclass()
    month_interval = NotImplemented
    n = NotImplemented

    def compute(self):
        if self.context > self.interval_end:
            return  # no need to compute: there will be no results
        self.current_month = gocept.month.IMonth(self.interval_start.date())
        # Adjust self.current_month to a multiple of self.month_interval:
        interval_len = len(
            self.current_month - gocept.month.IMonth(self.context.date()))
        # We have to substract 1 because a month interval with the same month
        # as start and end has a len of 1 but we need it zero based here:
        self.current_month += (interval_len - 1) % self.month_interval

        time = self.context.time()
        tz = self.context.tzinfo
        while True:
            result = next_date_of_same_weekday(
                self.context,
                tz.localize(datetime.combine(
                    self.current_month.firstOfMonth(), time)),
                self.n)
            try:
                if result.month != self.current_month.month:
                    continue  # result has swapped into next month
            finally:
                self.current_month += self.month_interval
            if result >= self.interval_end:
                break
            if result < self.context:
                continue
            yield result


class SameNthWeekdayFromBeginningInMonthBase(SameNthWeekdayInMonthBase):
    """Base class

    For recurrings on the same day nth weekday in month counting from the
    beginning of the month.
    """

    grok.baseclass()

    n_mapping = {0: _('1st'),
                 1: _('2nd'),
                 2: _('3rd'),
                 3: _('4th'),
                 4: _('5th')}

    @zope.cachedescriptors.property.Lazy
    def n(self):
        return int(math.ceil(self.context.day / 7.0)) - 1

    @property
    def info(self):
        return _(self.message_id,
                 mapping={'recurrence': self.n_mapping[self.n],
                          'weekday': self._weekday})


class MonthlyNthWeekday(SameNthWeekdayFromBeginningInMonthBase):
    """Recurring monthly on same recurrence of the weekday in the month as ...

    ... in `self.context`.
    """

    grok.name('nth weekday of month')
    weight = 20
    title = _('monthly, same weekday (e. g. each 3rd Sunday)')
    month_interval = 1
    message_id = _('${recurrence} ${weekday} every month')


class BiMonthlyNthWeekday(SameNthWeekdayFromBeginningInMonthBase):
    """Recurring monthly on same recurrence of the weekday in the month as ...

    ... in `self.context` but only every other month.
    """

    grok.name('nth weekday every other month')
    weight = 25
    title = _('every other month, same weekday '
              '(e. g. each 3rd Sunday in other month)')
    month_interval = 2
    message_id = _('${recurrence} ${weekday} every other month')


class SameNthWeekdayFromEndInMonthBase(SameNthWeekdayInMonthBase):
    """Base class for recurrings on the same day nth weekday in month ...

    ... counting from the end of the month.
    """

    grok.baseclass()

    n_mapping = {1: _('last'),
                 2: _('last but one'),
                 3: _('last but two'),
                 4: _('last but three'),
                 5: _('last but four')}

    @zope.cachedescriptors.property.Lazy
    def n_from_end(self):
        last_of_month = gocept.month.IMonth(self.context).lastOfMonth()
        return int(
            math.ceil((last_of_month.day - (self.context.day - 1)) / 7.0))

    @property
    def n(self):
        return recurrences_of_weekday_in_month(
            self.context, self.current_month) - self.n_from_end

    @property
    def info(self):
        return _(self.message_id,
                 mapping={'recurrence': self.n_mapping[self.n_from_end],
                          'weekday': self._weekday})


class MonthlyNthWeekdayFromEnd(SameNthWeekdayFromEndInMonthBase):
    """Recurring monthly on same recurrence of the weekday in the month as ...

    ... in `self.context`.
    """

    grok.name('nth weekday from end of month')
    weight = 21
    title = _('monthly, same weekday counted from the end of the month '
              '(e. g. each last but one Sunday)')
    month_interval = 1
    message_id = _('${recurrence} ${weekday} every month')


class BiMonthlyNthWeekdayFromEnd(SameNthWeekdayFromEndInMonthBase):
    """Recurring monthly on same recurrence of the weekday in the month as ...

    ... in `self.context` but only each other month.
    """

    grok.name('nth weekday from end of other month')
    weight = 26
    title = _('every other month on same weekday counted from the end of the '
              'month (e. g. each last but one Sunday every other month)')
    month_interval = 2
    message_id = _('${recurrence} ${weekday} every other month')



import pytz
from .interfaces import IRecurringDateTime
from .monthly import MonthlyNthWeekday, BiMonthlyNthWeekday
from .monthly import MonthlyNthWeekdayFromEnd, BiMonthlyNthWeekdayFromEnd
from .monthly import recurrences_of_weekday_in_month
from gocept.month import Month
from zope.interface.verify import verifyObject
import pytest


# Fixtures


@pytest.fixture('module')
def recurrence_start(DateTime):
    """Return default datetime when the recurrence should start."""
    return DateTime(2013, 3, 21, 21, 45)


@pytest.fixture('module')
def interval_start(DateTime):
    """Return the default start of the recurrence interval."""
    return DateTime(2014, 4, 1, 0)


@pytest.fixture('module')
def interval_end(DateTime):
    """Return the default end of the recurrence interval."""
    return DateTime(2014, 4, 30, 0)


# Tests


def test_monthly__recurrences_of_weekday_in_month__1(DateTime):
    """There are 4 recurrences of Monday in 7/2014."""
    assert (4 == recurrences_of_weekday_in_month(
        DateTime(2014, 8, 4), Month(7, 2014)))


def test_monthly__recurrences_of_weekday_in_month__2(DateTime):
    """There are 4 recurrences of Thursday in 6/2014."""
    assert (4 == recurrences_of_weekday_in_month(
        DateTime(2014, 8, 21), Month(6, 2014)))


def test_monthly__recurrences_of_weekday_in_month__3(DateTime):
    """There are 4 recurrences of Sunday in 7/2014."""
    assert (4 == recurrences_of_weekday_in_month(
        DateTime(2014, 8, 17), Month(7, 2014)))


def test_monthly__recurrences_of_weekday_in_month__4(DateTime):
    """There are 5 recurrences of Monday in 6/2014."""
    assert (5 == recurrences_of_weekday_in_month(
        DateTime(2014, 8, 4), Month(6, 2014)))


def test_monthly__recurrences_of_weekday_in_month__5(DateTime):
    """There are 5 recurrences of Thursday in 7/2014."""
    assert (5 == recurrences_of_weekday_in_month(
        DateTime(2014, 8, 21), Month(7, 2014)))


def test_monthly__recurrences_of_weekday_in_month__6(DateTime):
    """There are 5 recurrences of Sunday in 6/2014."""
    assert (5 == recurrences_of_weekday_in_month(
        DateTime(2014, 8, 17), Month(6, 2014)))


def test_monthly__MonthlyNthWeekday__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, MonthlyNthWeekday(today))


def test_monthly__MonthlyNthWeekday__info__1(info, recurrence_start):
    """It renders the weekday and the recurrence."""
    assert (u'3rd Thursday every month' ==
            info(MonthlyNthWeekday, recurrence_start))


def test_monthly__MonthlyNthWeekday____call____1(
        DateTime, interval_start, interval_end):
    """It returns an empty interval if the recurrence start is after the ...

    ...interval end.
    """
    dt = DateTime(2014, 5, 1, 21, 45)
    assert [] == list(MonthlyNthWeekday(dt)(interval_start, interval_end))


def test_monthly__MonthlyNthWeekday____call____2(
        DateTime, recurrence_start, interval_start):
    """It returns all nth of month in interval for same weekday."""
    end = DateTime(2014, 6, 30, 17)
    result = list(MonthlyNthWeekday(recurrence_start)(interval_start, end))
    assert [
        DateTime(2014, 4, 17, 21, 45),
        DateTime(2014, 5, 15, 21, 45),
        DateTime(2014, 6, 19, 21, 45)] == result
    assert recurrence_start.isoweekday() == result[0].isoweekday()


def test_monthly__MonthlyNthWeekday____call____3(DateTime, interval_start):
    """It does not start before the recurrence start date."""
    dt = DateTime(2014, 5, 4, 21, 45)
    end = DateTime(2014, 6, 30, 17)
    result = list(MonthlyNthWeekday(dt)(interval_start, end))
    assert [
        DateTime(2014, 5, 4, 21, 45),
        DateTime(2014, 6, 1, 21, 45)] == result


def test_monthly__MonthlyNthWeekday____call____4(
        DateTime, interval_start, interval_end):
    """`interval_end` does not belong to the interval."""
    dt = DateTime(2014, 4, 30, 0)
    assert [] == list(MonthlyNthWeekday(dt)(interval_start, interval_end))


def test_monthly__MonthlyNthWeekday____call____5(
        DateTime, interval_start, interval_end):
    """`interval_start` belongs to the interval."""
    dt = DateTime(2014, 4, 1, 0)
    assert ([DateTime(2014, 4, 1, 0)] ==
            list(MonthlyNthWeekday(dt)(interval_start, interval_end)))


def test_monthly__MonthlyNthWeekday____call____6(DateTime):
    """It does not swap into the next month if the month do not have a ...

    ... fifth week.
    """
    dt = DateTime(2014, 5, 31, 0)
    start = DateTime(2014, 5, 1, 0)
    end = DateTime(2014, 8, 31, 0)
    assert [
        DateTime(2014, 5, 31, 0),
        DateTime(2014, 8, 30, 0)] == list(MonthlyNthWeekday(dt)(start, end))


def test_monthly__MonthlyNthWeekday____call____7(DateTime):
    """If the context has a timezone which has DST it is respected.

    So the local time does not change if DST switches.
    DST ... daylight saving time
    """
    tz_berlin = pytz.timezone('Europe/Berlin')
    adapter = MonthlyNthWeekday(DateTime(2016, 3, 24, 12, tzinfo=tz_berlin))
    # At 2016-03-27 DST starts in Europe/Berlin
    result = list(adapter(DateTime(2016, 3, 24, 0), DateTime(2016, 5, 1, 0)))
    assert ([DateTime(2016, 3, 24, 12, tzinfo=tz_berlin),
             DateTime(2016, 4, 28, 12, tzinfo=tz_berlin)] == result)
    # So the time in UTC changes to keep it the same in local time:
    assert ([DateTime(2016, 3, 24, 11),
             DateTime(2016, 4, 28, 10)] ==
            [pytz.utc.normalize(x) for x in result])


def test_monthly__BiMonthlyNthWeekday__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, BiMonthlyNthWeekday(today))


def test_monthly__BiMonthlyNthWeekday__info__1(info, recurrence_start):
    """It renders the weekday and the recurrence."""
    assert (u'3rd Thursday every other month' ==
            info(BiMonthlyNthWeekday, recurrence_start))


def test_monthly__BiMonthlyNthWeekday____call____1(
        DateTime, recurrence_start, interval_start):
    """It returns all nth of month in the interval every other month."""
    end = DateTime(2014, 7, 31)
    result = list(BiMonthlyNthWeekday(recurrence_start)(interval_start, end))
    assert [
        DateTime(2014, 5, 15, 21, 45),
        DateTime(2014, 7, 17, 21, 45)] == result
    assert recurrence_start.isoweekday() == result[0].isoweekday()


def test_monthly__MonthlyNthWeekdayFromEnd__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, MonthlyNthWeekdayFromEnd(today))


def test_monthly__MonthlyNthWeekdayFromEnd__info__1(info, recurrence_start):
    """It renders the weekday and the recurrence."""
    assert (u'last but one Thursday every month' ==
            info(MonthlyNthWeekdayFromEnd, recurrence_start))


def test_monthly__MonthlyNthWeekdayFromEnd____call____1(
        DateTime, interval_start, interval_end):
    """It returns an empty interval if the recurrence start date is after ...

    ... the interval end.
    """
    dt = DateTime(2014, 5, 1, 21, 45)
    assert ([] ==
            list(MonthlyNthWeekdayFromEnd(dt)(interval_start, interval_end)))


def test_monthly__MonthlyNthWeekdayFromEnd____call____2(
        DateTime, recurrence_start, interval_start):
    """It returns all nth from end of the month in the interval for same ...

    ... weekday.
    """
    end = DateTime(2014, 6, 30, 17)
    result = list(MonthlyNthWeekdayFromEnd(
        recurrence_start)(interval_start, end))
    assert [
        DateTime(2014, 4, 17, 21, 45),
        DateTime(2014, 5, 22, 21, 45),
        DateTime(2014, 6, 19, 21, 45)] == result
    assert recurrence_start.isoweekday() == result[0].isoweekday()


def test_monthly__MonthlyNthWeekdayFromEnd____call____3(
        DateTime, interval_start):
    """It does not start before the recurrence start date."""
    dt = DateTime(2014, 5, 4, 21, 45)
    end = DateTime(2014, 6, 30, 17)
    result = list(MonthlyNthWeekdayFromEnd(dt)(interval_start, end))
    assert [
        DateTime(2014, 5, 4, 21, 45),
        DateTime(2014, 6, 8, 21, 45)] == result


def test_monthly__MonthlyNthWeekdayFromEnd____call____4(
        DateTime, interval_start, interval_end):
    """`interval_end` does not belong to the interval."""
    dt = DateTime(2014, 4, 30, 0)
    assert ([] ==
            list(MonthlyNthWeekdayFromEnd(dt)(interval_start, interval_end)))


def test_monthly__MonthlyNthWeekdayFromEnd____call____5(
        DateTime, interval_start, interval_end):
    """`interval_start` belongs to the interval."""
    dt = DateTime(2014, 4, 1, 0)
    assert ([DateTime(2014, 4, 1, 0)] ==
            list(MonthlyNthWeekdayFromEnd(dt)(interval_start, interval_end)))


def test_monthly__MonthlyNthWeekdayFromEnd____call____6(DateTime):
    """It does not swap into the next month if the month do not have a ...

    ...fifth week.
    """
    dt = DateTime(2014, 5, 3, 0)
    start = DateTime(2014, 5, 1, 0)
    end = DateTime(2014, 8, 31, 0)
    result = list(MonthlyNthWeekdayFromEnd(dt)(start, end))
    assert [
        DateTime(2014, 5, 3, 0),
        DateTime(2014, 8, 2, 0)] == result


def test_monthly__BiMonthlyNthWeekdayFromEnd__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, BiMonthlyNthWeekdayFromEnd(today))


def test_monthly__BiMonthlyNthWeekdayFromEnd__info__1(info, recurrence_start):
    """It renders the weekday and the recurrence."""
    assert (u'last but one Thursday every other month' ==
            info(BiMonthlyNthWeekdayFromEnd, recurrence_start))


def test_monthly__BiMonthlyNthWeekdayFromEnd____call____1(
        DateTime, recurrence_start, interval_start):
    """It returns all nth from end of every other month in the interval for ...

    ... the same weekday.
    """
    end = DateTime(2014, 7, 31, 17)
    result = list(BiMonthlyNthWeekdayFromEnd(
        recurrence_start)(interval_start, end))
    assert [
        DateTime(2014, 5, 22, 21, 45),
        DateTime(2014, 7, 24, 21, 45)] == result
    assert recurrence_start.isoweekday() == result[0].isoweekday()



import datetime
import icemac.recurrence
import pytest
import pytz
import zope.configuration.xmlconfig
import zope.globalrequest
import zope.i18n
import zope.publisher.browser


@pytest.fixture(scope='session', autouse=True)
def zcmlS():
    """Load ZCML on session scope."""
    zope.configuration.xmlconfig.file('configure.zcml', icemac.recurrence)


@pytest.fixture('session')
def today():
    """Fixture returning the current day."""
    return datetime.date.today()


@pytest.fixture(scope='session')
def DateTime():
    """Fixture to ease handling of datetime objects."""
    return DateTimeClass()


class DateTimeClass:
    """Helper class to create and format datetime objects."""

    @staticmethod
    def __call__(*args, **kw):
        """Create a datetime object.

        `*args` ... time tuple parts
                    If no `args` are given the current time is returned.
        `**kw` ... the only allowed key is `tzinfo`.
                   If `tzinfo` is None, UTC is used instead.

        """
        tzinfo = kw.pop('tzinfo', pytz.utc)
        assert not kw  # make sure there are no kw left
        assert args  # make sure there is at least one argument
        return tzinfo.localize(datetime.datetime(*args))


@pytest.fixture('session')
def info():
    """Call `info` on a recurrence adapter and translate the result."""
    def info(adapter, datetime):
        request = zope.publisher.browser.TestRequest(HTTP_ACCEPT_LANGUAGE='en')
        # We have to set this request globally as otherwise the localization of
        # the weekdays returns only numbers:
        zope.globalrequest.setRequest(request)
        return zope.i18n.translate(adapter(datetime).info, request)
    return info



from .interfaces import IRecurringDateTime
import zope.component


def get_recurring(datetime, period):
    """Convenience function to get the recurring adapter named `period`."""
    return zope.component.getAdapter(datetime, IRecurringDateTime, name=period)


def get_recurrences(datetime, period, interval_start, interval_end):
    """Convenience function

    Goal: Get an interable of datetime objects of recurrences of `period`
    within the interval.

    period ... string, name of an adapter, see below
    interval_start ... date, part of the interval
    interval_end ... date, _not_ part of the interval

    """
    return get_recurring(datetime, period)(interval_start, interval_end)



from .interfaces import IRecurringDateTime
from datetime import timedelta, datetime
import grokcore.component as grok
import zope.interface.common.idatetime


ONE_DAY = timedelta(days=1)


def _get_isoweekday_difference(date1, date2):
    """Difference of isoweekdays between days."""
    days = 7 - (date1.isoweekday() - date2.isoweekday())
    if days >= 7:
        days -= 7
    return days


def next_date_of_same_weekday(wd_src, base_date, additional_weeks=0):
    """Compute next day with the same weekday as `wd_src` from `base_date` on.

    If `additional_weeks` is not zero, its number of weeks are added
    afterwards.
    If `wd_src` and `base_date` have the same weekday `base_date` is returned.

    """
    add_days = _get_isoweekday_difference(base_date, wd_src)
    return base_date + (add_days + additional_weeks * 7) * ONE_DAY


class RecurringDateTime(grok.Adapter):
    """Base class for recurring datestimes."""

    grok.context(zope.interface.common.idatetime.IDateTime)
    grok.implements(IRecurringDateTime)
    grok.baseclass()

    def __call__(self, interval_start, interval_end):
        self.interval_start = interval_start
        self.interval_end = interval_end
        return self.compute()

    def compute(self):
        raise NotImplementedError('Implement in subclass!')

    @property
    def _weekday(self):
        request = zope.globalrequest.getRequest()
        weekday = self.context.isoweekday()
        if request is not None:
            calendar = request.locale.dates.calendars['gregorian']
            return calendar.getDayNames()[weekday - 1]
        return weekday


class StaticIntervalBase(RecurringDateTime):
    """Base class for recurrences of a fix interval e. g. 1 day or 1 week."""

    grok.baseclass()
    interval = NotImplemented

    def _get_start_date(self):
        raise NotImplementedError('Implement in subclass!')

    def compute(self):
        if self.interval_start <= self.context:
            current_date = self.context
        else:
            current_date = self._get_start_date()
        time = self.context.time()
        tz = self.context.tzinfo
        while current_date < self.interval_end:
            yield tz.localize(datetime.combine(current_date, time))
            current_date += self.interval



from .base import RecurringDateTime, StaticIntervalBase
from .base import next_date_of_same_weekday
import pytest


def test_base__next_date_of_same_weekday__1(DateTime):
    """Weekday of `wd_src` smaller than weekday of base date."""
    assert (DateTime(2014, 7, 28, 10) == next_date_of_same_weekday(
        DateTime(2014, 7, 21, 10), DateTime(2014, 7, 23, 10)))


def test_base__next_date_of_same_weekday__2(DateTime):
    """Weekday of `wd_src` greater than weekday of base date."""
    # Weekday of 2014-07-20 is 7 (Sunday)
    assert (DateTime(2014, 7, 27, 10) == next_date_of_same_weekday(
        DateTime(2014, 7, 20, 10), DateTime(2014, 7, 23, 10)))


def test_base__next_date_of_same_weekday__3(DateTime):
    """Weekday of `wd_src` equal to weekday of base date."""
    assert (DateTime(2014, 7, 17, 10) == next_date_of_same_weekday(
        DateTime(2014, 7, 10, 10), DateTime(2014, 7, 17, 10)))


def test_base__next_date_of_same_weekday__4(DateTime):
    """`wd_src` equal to base date."""
    dt = DateTime(2014, 7, 23, 17)
    assert (dt == next_date_of_same_weekday(dt, dt))


def test_base__next_date_of_same_weekday__5(DateTime):
    """Additional whole weeks can be added."""
    assert (DateTime(2014, 9, 9, 15) == next_date_of_same_weekday(
        DateTime(2014, 9, 2, 15), DateTime(2014, 9, 1, 15), 1))


def test_base__RecurringDateTime__compute__1():
    """It needs to be implemented by child classes."""
    instance = RecurringDateTime(None)
    with pytest.raises(NotImplementedError):
        instance.compute()


def test_base__RecurringDateTime___weekday__1(DateTime):
    """It is the week day number if there is no request."""
    instance = RecurringDateTime(DateTime(2016, 2, 20))  # Saturday
    assert 6 == instance._weekday


def test_base__StaticIntervalBase___get_start_date__1():
    """It needs to be implemented by child classes."""
    instance = StaticIntervalBase(None)
    with pytest.raises(NotImplementedError):
        instance._get_start_date()



from .recurrence import get_recurring, get_recurrences
from .weekly import Weekly
from zope.component import ComponentLookupError
import collections
import pytest


def test_recurrence__get_recurring__1(DateTime):
    """It returns the named recurrence adapter."""
    assert Weekly == get_recurring(DateTime(2016, 2, 6), 'weekly').__class__


def test_recurrence__get_recurring__2(DateTime):
    """It raises a TypeError if there is no adapter registered for the name."""
    with pytest.raises(ComponentLookupError):
        get_recurring(DateTime(2016, 2, 6), 'foobar')


def test_recurrence__get_recurrences__1(DateTime):
    """It returns an iterable of datetime objects."""
    dt = DateTime(2016, 2, 6, 10)
    start = DateTime(2016, 3, 1)
    end = DateTime(2016, 6, 30)
    result = get_recurrences(
        dt, 'nth weekday from end of other month', start, end)
    assert isinstance(result, collections.Iterable)
    assert [DateTime(2016, 4, 9, 10),
            DateTime(2016, 6, 4, 10)] == list(result)



import zope.interface


class IRecurringDateTime(zope.interface.Interface):
    """Recurring of a datetime.

    Period and base datetime are defined in class implementing the interface.
    """

    title = zope.interface.Attribute('Display title in RecurrencePeriodSource')
    weight = zope.interface.Attribute(
        'RecurrencePeriodSource uses `weight` to sort.')
    info = zope.interface.Attribute(
        'Information about recurrence period e. g. `every sunday`.')

    def __call__(interval_start, interval_end):
        """Iterable of recurrences of base datetime in the interval.

        interval_start, interval_end ... `datetime.date` objects

        """



from icemac.recurrence.i18n import _
import grokcore.component as grok
from .base import StaticIntervalBase, ONE_DAY


class Daily(StaticIntervalBase):
    """Recurring each day."""

    grok.name('daily')
    weight = 5
    interval = ONE_DAY
    title = _('daily')
    info = _('each day')

    def _get_start_date(self):
        return self.interval_start



# -*- coding: utf-8 -*-
import os.path
import setuptools


def read(*path_elements):
    """Read file."""
    return file(os.path.join(*path_elements)).read()


version = '1.2'
long_description = '\n\n'.join([
    read('README.rst'),
    read('CHANGES.rst'),
])

setuptools.setup(
    name='icemac.recurrence',
    version=version,
    description="Compute recurrences of events",
    long_description=long_description,
    keywords='calendar event recurring recurrence datetime',
    author='Michael Howitz',
    author_email='icemac@gmx.net',
    download_url='http://pypi.python.org/pypi/icemac.recurrence',
    url='https://bitbucket.org/icemac/icemac.recurrence',
    license='ZPL 2.1',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved',
        'License :: OSI Approved :: Zope Public License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['icemac'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'gocept.month >= 1.4.dev',
        'grokcore.component >= 2.5.1.dev1',
        'setuptools',
        'zope.cachedescriptors',
        'zope.component',
        'zope.globalrequest',
        'zope.i18nmessageid',
        'zope.interface',
    ],
    extras_require=dict(
        test=[
            'zope.configuration',
        ]),
)



from .interfaces import IRecurringDateTime
from .weekly import Weekly, BiWeekly
from zope.interface.verify import verifyObject
import pytest
import pytz


# Fixtures


@pytest.fixture('module')
def recurrence_start(DateTime):
    """Return default datetime when the recurrence should start."""
    return DateTime(2013, 5, 3, 21, 45)


@pytest.fixture('module')
def interval_start(DateTime):
    """Return the default start of the recurrence interval."""
    return DateTime(2014, 4, 1)


@pytest.fixture('module')
def interval_end(DateTime):
    """Return the default end of the recurrence interval."""
    return DateTime(2014, 4, 30)


# Tests


def test_weekly__Weekly__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, Weekly(today))


def test_weekly__Weekly__info__1(info, recurrence_start):
    """It renders the weekday."""
    assert u'Friday every week' == info(Weekly, recurrence_start)


def test_weekly__Weekly____call____1(
        DateTime, recurrence_start, interval_start, interval_end):
    """It returns all dates in the interval for same weekday."""
    result = list(Weekly(recurrence_start)(interval_start, interval_end))
    assert [
        DateTime(2014, 4, 4, 21, 45),
        DateTime(2014, 4, 11, 21, 45),
        DateTime(2014, 4, 18, 21, 45),
        DateTime(2014, 4, 25, 21, 45)] == result
    assert recurrence_start.isoweekday() == result[0].isoweekday()


def test_weekly__Weekly____call____2(DateTime, interval_start, interval_end):
    """The returned values do not start before the recurrence start."""
    dt = DateTime(2014, 4, 18, 21, 45)
    result = list(Weekly(dt)(interval_start, interval_end))
    assert [
        DateTime(2014, 4, 18, 21, 45),
        DateTime(2014, 4, 25, 21, 45)] == result
    assert dt.isoweekday() == result[0].isoweekday()


def test_weekly__Weekly____call____3(DateTime, interval_start, interval_end):
    """It returns an empty iterable if the recurrence should start after ...

    ... `interval_end`.
    """
    assert ([] == list(Weekly(
        DateTime(2014, 5, 1, 21, 45))(interval_start, interval_end)))


def test_weekly__Weekly____call____4(DateTime, recurrence_start):
    """`interval_end` does not belong to the interval."""
    assert ([] == list(Weekly(
        recurrence_start)(DateTime(2014, 4, 24), DateTime(2014, 4, 25))))


def test_weekly__Weekly____call____5(DateTime, recurrence_start):
    """`interval_start` belongs to the interval."""
    assert ([DateTime(2014, 4, 4, 21, 45)] == list(Weekly(
        recurrence_start)(DateTime(2014, 4, 4), DateTime(2014, 4, 5))))


def test_weekly__Weekly____call____6(DateTime):
    """If the context has a timezone which has DST it is respected.

    So the local time does not change if DST switches.
    DST ... daylight saving time
    """
    tz_berlin = pytz.timezone('Europe/Berlin')
    adapter = Weekly(DateTime(2016, 3, 24, 12, tzinfo=tz_berlin))
    # At 2016-03-27 DST starts in Europe/Berlin
    result = list(adapter(DateTime(2016, 3, 24, 0), DateTime(2016, 4, 1, 0)))
    assert ([DateTime(2016, 3, 24, 12, tzinfo=tz_berlin),
             DateTime(2016, 3, 31, 12, tzinfo=tz_berlin)] == result)
    # So the time in UTC changes to keep it the same in local time:
    assert ([DateTime(2016, 3, 24, 11),
             DateTime(2016, 3, 31, 10)] ==
            [pytz.utc.normalize(x) for x in result])


def test_weekly__BiWeekly__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, BiWeekly(today))


def test_weekly__BiWeekly____call____1(
        DateTime, recurrence_start, interval_start, interval_end):
    """returns_all_dates_in_interval_for_same_weekday_every_other_week."""
    result = list(BiWeekly(recurrence_start)(interval_start, interval_end))
    assert [
        DateTime(2014, 4, 4, 21, 45),
        DateTime(2014, 4, 18, 21, 45)] == result
    assert recurrence_start.isoweekday() == result[0].isoweekday()


def test_weekly__BiWeekly__info__1(info, recurrence_start):
    """It renders the weekday."""
    assert u'Friday every other week' == info(BiWeekly, recurrence_start)



from .base import RecurringDateTime
from icemac.recurrence.i18n import _
import grokcore.component as grok


def add_years(date, years):
    """Add `years` number of years to date."""
    try:
        return date.replace(year=date.year + years)
    except ValueError:
        # Handle 29th of February as 28th in non-leap years:
        return date.replace(year=date.year + years, day=date.day - 1)


class Yearly(RecurringDateTime):
    """Recurring on the same date each year."""

    grok.name('yearly')
    weight = 100
    title = _('yearly (e. g. 24th of December)')

    def compute(self):
        if self.context > self.interval_end:
            return  # no need to compute: there will be no results
        date = self.context
        index = 0
        # Find first date after interval_start:
        while date < self.interval_start:
            index += 1
            date = add_years(self.context, index)
        # Yield dates in the interval:
        while date < self.interval_end:
            yield date
            index += 1
            date = add_years(self.context, index)

    @property
    def info(self):
        return _('${date} every year',
                 mapping={'date': self.context.strftime('%d.%m.')})



from zope.interface.verify import verifyObject
from .interfaces import IRecurringDateTime
from .daily import Daily


def test_daily__Daily__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, Daily(today))


def test_daily__Daily____call____1(DateTime):
    """It returns all dates in the interval."""
    dt = DateTime(2013, 5, 3, 21, 45)
    start = DateTime(2014, 4, 1)
    end = DateTime(2014, 4, 4)
    assert [
        DateTime(2014, 4, 1, 21, 45),
        DateTime(2014, 4, 2, 21, 45),
        DateTime(2014, 4, 3, 21, 45)] == list(Daily(dt)(start, end))


def test_daily__Daily__info__1(info):
    """It renders a static string."""
    assert u'each day' == info(Daily, None)



from .interfaces import IRecurringDateTime
from .yearly import Yearly
from zope.interface.verify import verifyObject
import pytest


# Fixtures

@pytest.fixture('module')
def recurrence_start(DateTime):
    """Return default datetime when the recurrence should start."""
    return DateTime(2013, 12, 24, 15)


@pytest.fixture('module')
def interval_start(DateTime):
    """Return the default start of the recurrence interval."""
    return DateTime(2014, 1, 1)


@pytest.fixture('module')
def interval_end(DateTime):
    """Return the default end of the recurrence interval."""
    return DateTime(2014, 12, 31)


# Tests


def test_yearly__Yearly__1(today):
    """It fulfills the `IRecurringDateTime` interface."""
    assert verifyObject(IRecurringDateTime, Yearly(today))


def test_yearly__Yearly__info__1(info, recurrence_start):
    """It renders the recurrence day."""
    assert (u'24.12. every year' == info(Yearly, recurrence_start))


def test_yearly__Yearly____call____1(DateTime, recurrence_start):
    """It returns all dates in the interval with the same day and month."""
    start = DateTime(2012, 1, 1)
    end = DateTime(2015, 1, 1)
    assert ([
        DateTime(2013, 12, 24, 15),
        DateTime(2014, 12, 24, 15)] ==
        list(Yearly(recurrence_start)(start, end)))


def test_yearly__Yearly____call____2(
        DateTime, recurrence_start, interval_start):
    """It returns an empty interval if the recurrence start is after the ...

    ... interval end.
    """
    end = DateTime(2012, 5, 1)
    assert [] == list(Yearly(recurrence_start)(interval_start, end))


def test_yearly__Yearly____call____3(
        DateTime, recurrence_start, interval_start):
    """`interval_end` does not belong to the interval."""
    end = DateTime(2014, 12, 24, 15)
    assert [] == list(Yearly(recurrence_start)(interval_start, end))


def test_yearly__Yearly____call____4(DateTime, recurrence_start, interval_end):
    """`interval_start` belongs to the interval."""
    start = DateTime(2014, 12, 24, 15)
    assert ([DateTime(2014, 12, 24, 15)] ==
            list(Yearly(recurrence_start)(start, interval_end)))


def test_yearly__Yearly____call____5(DateTime):
    """It handles the 29th of february as 28th in a non-leap year."""
    dt = DateTime(2008, 2, 29, 15)
    start = DateTime(2011, 1, 1)
    end = DateTime(2013, 1, 1)
    assert [
        DateTime(2011, 2, 28, 15),
        DateTime(2012, 2, 29, 15)] == list(Yearly(dt)(start, end))



from .base import StaticIntervalBase, next_date_of_same_weekday
from datetime import timedelta
from icemac.recurrence.i18n import _
import grokcore.component as grok


ONE_WEEK = timedelta(days=7)
TWO_WEEKS = timedelta(days=14)


class SameWeekdayBase(StaticIntervalBase):
    """Base class for recurrences on the same weekday."""

    grok.baseclass()

    def _get_start_date(self):
        return next_date_of_same_weekday(self.context, self.interval_start)


class Weekly(SameWeekdayBase):
    """Recurring weekly on the same weekday."""

    grok.name('weekly')
    weight = 10
    title = _('weekly, same weekday (e. g. each Friday)')
    interval = ONE_WEEK

    @property
    def info(self):
        return _('${weekday} every week',
                 mapping={'weekday': self._weekday})


class BiWeekly(SameWeekdayBase):
    """Recurring biweekly on the same weekday."""

    grok.name('biweekly')
    weight = 11
    title = _('every other week, same weekday (e. g. each second Friday)')
    interval = TWO_WEEKS

    @property
    def info(self):
        return _('${weekday} every other week',
                 mapping={'weekday': self._weekday})
